from hitchbuild.build import HitchBuild
from copy import copy


class BuildBundle(object):
    def __init__(self, build_path):
        self._build_path = build_path
        self._sqlite_filename = build_path.build.joinpath("builddb.sqlite")
        self._builds = {}

    def __setitem__(self, name, build):
        assert isinstance(name, str), "name must be a string"
        assert isinstance(build, HitchBuild), "must assign a build of type HitchBuild"
        self._builds[name] = build.with_name(name)\
                                  .with_db(self._sqlite_filename)\
                                  .in_path(self._build_path)

    def __getitem__(self, name):
        return self._builds[name]

    def ensure_built(self):
        build_triggered = True

        while build_triggered:
            build_triggered = False
            for build in self._builds.values():
                if build.ensure_built():
                    build_triggered = True

    def manually_trigger(self, *names):
        new_bundle = copy(self)
        for name in names:
            new_bundle[name] = new_bundle[name].manually_triggered()
        return new_bundle



from hitchbuild.monitor import Monitor
from hitchbuild.condition import Always
from path import Path
from copy import copy


class HitchBuild(object):
    def __init__(self):
        pass

    def exists(self):
        raise NotImplemented()

    def build(self):
        raise NotImplementedError("build method must be implemented")

    def trigger(self):
        return Always()

    @property
    def monitor(self):
        return Monitor(self.name, self.build_database)

    def with_name(self, name):
        new_build = copy(self)
        new_build._name = name
        return new_build

    @property
    def build_path(self):
        return self._build_path

    @property
    def name(self):
        if hasattr(self, '_name'):
            return self._name
        else:
            return type(self).__name__

    def with_build_path(self, path):
        new_build = copy(self)
        new_build._build_path = Path(path)
        return new_build

    def with_default_build_path(self, path):
        if hasattr(self, '_build_path'):
            return self
        else:
            return self.with_build_path(path)

    def with_db(self, sqlite_filename):
        new_build = copy(self)
        new_build._sqlite_filename = sqlite_filename
        return new_build

    @property
    def build_database(self):
        if hasattr(self, '_sqlite_filename'):
            return self._sqlite_filename
        else:
            return self.build_path / "builddb.sqlite"

    def triggered(self):
        new_build = copy(self)
        new_build._manually_triggered = True
        return new_build

    @property
    def was_manually_triggered(self):
        if hasattr(self, '_manually_triggered'):
            return self._manually_triggered
        else:
            return False

    def as_dependency(self, build):
        dependent_build = build
        if hasattr(self, '_dependencies'):
            self._dependencies.append(dependent_build)
        else:
            self._dependencies = [dependent_build, ]
        return dependent_build

    def ensure_built(self):
        dependency_triggered = False

        if hasattr(self, '_dependencies'):
            for dependency in self._dependencies:
                if dependency.with_default_build_path(self.build_path).ensure_built():
                    dependency_triggered = True

        trigger_check = self.trigger().check_all()

        triggered = (
            trigger_check or
            self.monitor.last_run_had_exception or
            self.was_manually_triggered or
            dependency_triggered
        )

        if triggered:
            with self.monitor.context_manager():
                self.build()

        model = self.monitor.build_model
        model.was_triggered_on_last_run = triggered
        model.save()

        return triggered

    def __repr__(self):
        return """HitchBuild("{0}")""".format(self.name)



from hitchstory import StoryCollection, StorySchema, BaseEngine, exceptions
from hitchstory import validate, expected_exception
from hitchrun import hitch_maintenance, expected, DIR
from pathquery import pathq
from strictyaml import MapPattern, Optional, Str, Int
from commandlib import python, Command
import hitchpython
import strictyaml
import hitchtest
from hitchrunpy import ExamplePythonCode
from templex import Templex


class Engine(BaseEngine):
    schema = StorySchema(
        given={
            Optional("python version"): Str(),
            Optional("build.py"): Str(),
            Optional("setup"): Str(),
            Optional("files"): MapPattern(Str(), Str()),
        },
        params={
            "python version": Str(),
        },
        info={
            Optional("about"): Str(),
        },
    )

    def __init__(self, paths, settings):
        self.path = paths
        self.settings = settings

    def set_up(self):
        self.path.state = self.path.gen.joinpath("state")

        if self.path.state.exists():
            self.path.state.rmtree(ignore_errors=True)
        self.path.state.mkdir()

        for filename, content in self.given.get("files", {}).items():
            filepath = self.path.state.joinpath(filename)
            if not filepath.dirname().exists():
                filepath.dirname().makedirs()
            filepath.write_text(content)

        for filename in ["build.py", "sourcefile.txt", ]:
            if filename in self.given:
                filepath = self.path.state.joinpath(filename)
                if not filepath.dirname().exists():
                    filepath.dirname().mkdir()
                filepath.write_text(str(self.given[filename]))

        self.path.key.joinpath("code_that_does_things.py").copy(self.path.state)

        self.python_package = hitchpython.PythonPackage(
            self.given.get('python_version', self.given['python version'])
        )
        self.python_package.build()

        self.pip = self.python_package.cmd.pip
        self.python = self.python_package.cmd.python

        with hitchtest.monitor([self.path.key.joinpath("debugrequirements.txt")]) as changed:
            if changed:
                self.pip("install", "-r", "debugrequirements.txt").in_dir(self.path.key).run()

        with hitchtest.monitor(
            pathq(self.path.project.joinpath("hitchbuild")).ext("py")
        ) as changed:
            if changed:
                self.pip("uninstall", "hitchbuild", "-y").ignore_errors().run()
                self.pip("install", ".").in_dir(self.path.project).run()

    def run_code(self, code):
        ExamplePythonCode(code).with_setup_code(self.given.get('setup', ''))\
                               .run(self.path.state, self.python)

    def exception_raised_with(self, code, exception_type, message):
        result = ExamplePythonCode(code).with_setup_code(self.given.get('setup', ''))\
                                        .expect_exceptions()\
                                        .run(self.path.state, self.python)
        result.exception_was_raised(exception_type, message)

    def run(self, command):
        self.ipython_step_library.run(command)

    def touch_file(self, filename):
        self.path.state.joinpath(filename).write_text("\nfile touched!", append=True)

    @expected_exception(AssertionError)
    def file_exists(self, filename):
        assert self.path.state.joinpath(filename).exists(), \
           "{0} does not exist".format(filename)

    @expected_exception(FileNotFoundError)
    @expected_exception(AssertionError)
    def file_contents_will_be(self, filename, text=None):
        try:
            Templex(self.path.state.joinpath(filename).text()).assert_match(text)
        except AssertionError:
            if self.settings.get("overwrite artefacts"):
                self.current_step.update(
                    text=self.path.state.joinpath(filename).text()
                )
            else:
                raise

    @validate(seconds=Int())
    def sleep(self, seconds):
        import time
        time.sleep(int(seconds))

    def on_success(self):
        if self.settings.get("overwrite artefacts"):
            self.new_story.save()


@expected(strictyaml.exceptions.YAMLValidationError)
@expected(exceptions.HitchStoryException)
def bdd(*keywords):
    """
    Run story matching keywords specified.
    """
    StoryCollection(
        pathq(DIR.key).ext("story"),
        Engine(DIR, {"overwrite artefacts": False})
    ).shortcut(*keywords).play()


@expected(strictyaml.exceptions.YAMLValidationError)
@expected(exceptions.HitchStoryException)
def rbdd(*keywords):
    """
    Run story matching keywords specified.
    """
    StoryCollection(
        pathq(DIR.key).ext("story"),
        Engine(DIR, {"overwrite artefacts": True})
    ).shortcut(*keywords).play()


def regression():
    """
    Run all stories.
    """
    lint()
    StoryCollection(
        pathq(DIR.key).ext("story"), Engine(DIR, {})
    ).ordered_by_name().play()


def hitch(*args):
    """
    Use 'h hitch --help' to get help on these commands.
    """
    hitch_maintenance(*args)


def lint():
    """
    Lint all code.
    """
    python("-m", "flake8")(
        DIR.project.joinpath("hitchbuild"),
        "--max-line-length=100",
        "--exclude=__init__.py",
    ).run()
    python("-m", "flake8")(
        DIR.key.joinpath("key.py"),
        "--max-line-length=100",
    ).run()
    print("Lint success!")


def deploy(version):
    """
    Deploy to pypi as specified version.
    """
    NAME = "hitchbuild"
    git = Command("git").in_dir(DIR.project)
    version_file = DIR.project.joinpath("VERSION")
    old_version = version_file.bytes().decode('utf8')
    if version_file.bytes().decode("utf8") != version:
        DIR.project.joinpath("VERSION").write_text(version)
        git("add", "VERSION").run()
        git("commit", "-m", "RELEASE: Version {0} -> {1}".format(
            old_version,
            version
        )).run()
        git("push").run()
        git("tag", "-a", version, "-m", "Version {0}".format(version)).run()
        git("push", "origin", version).run()
    else:
        git("push").run()

    # Set __version__ variable in __init__.py, build sdist and put it back
    initpy = DIR.project.joinpath(NAME, "__init__.py")
    original_initpy_contents = initpy.bytes().decode('utf8')
    initpy.write_text(
        original_initpy_contents.replace("DEVELOPMENT_VERSION", version)
    )
    python("setup.py", "sdist").in_dir(DIR.project).run()
    initpy.write_text(original_initpy_contents)

    # Upload to pypi
    python(
        "-m", "twine", "upload", "dist/{0}-{1}.tar.gz".format(NAME, version)
    ).in_dir(DIR.project).run()



import sys


class ExampleException(Exception):
    """
    This is a demonstration exception's docstring.

    It spreads across multiple lines.
    """
    pass


def should_run(which):
    with open("should{0}.txt".format(which), "w") as handle:
        handle.write("ran!")


def should_not_run():
    raise RuntimeError("This shouldn't have happened")


def raise_example_exception(text=""):
    raise ExampleException(text)


def output(contents):
    with open("output.txt", 'w') as handle:
        handle.write("{0}\n".format(contents))


def append(contents):
    with open("output.txt", 'a') as handle:
        handle.write("{0}\n".format(contents))


def reticulate_splines():
    with open("splines_reticulated.txt", 'w') as handle:
        handle.write("{0}\n".format("splines_reticulated"))


def kick_llamas_ass():
    with open("kicked_llamas_ass.txt", 'w') as handle:
        handle.write("{0}\n".format("kicked_llamas_ass"))



from path import Path


class BuildPath(object):
    def __init__(self, **paths):
        assert "build" in paths, "build=pathname must be set at a minimum"
        for name, path in paths.items():
            assert Path(path), "must be valid path"
            setattr(self, name, Path(path).abspath())



from peewee import ForeignKeyField, CharField, FloatField, BooleanField, DateTimeField
from peewee import SqliteDatabase, Model
from datetime import timedelta as python_timedelta
from datetime import datetime as python_datetime
from hitchbuild import condition


class BuildContextManager(object):
    def __init__(self, monitor):
        self._monitor = monitor

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        model = self._monitor.build_model
        model.exception_raised = traceback is not None
        model.last_run = python_datetime.now()
        model.save()


class Monitor(object):
    """
    The HitchBuild Monitor keeps track of things which
    determine whether to trigger a rebuild.
    """
    def __init__(self, name, sqlite_filename):
        self._name = name

        class BaseModel(Model):
            class Meta:
                database = SqliteDatabase(str(sqlite_filename))

        class Build(BaseModel):
            name = CharField(primary_key=True)
            exception_raised = BooleanField()
            last_run = DateTimeField(null=True)
            was_triggered_on_last_run = BooleanField(null=True)

        class File(BaseModel):
            build = ForeignKeyField(Build)
            filename = CharField(max_length=640)
            last_modified = FloatField()

        if not Build.table_exists():
            Build.create_table()
        if not File.table_exists():
            File.create_table()

        self.BaseModel = BaseModel
        self.File = File
        self.Build = Build

    @property
    def build_model(self):
        if self.Build.filter(name=self._name).first() is None:
            model = self.Build(
                name=self._name,
                exception_raised=False,
                last_run=None,
                was_triggered_on_last_run=None
            )
            model.save(force_insert=True)
        else:
            model = self.Build.filter(name=self._name).first()
        return model

    @property
    def last_run_had_exception(self):
        return self.build_model.exception_raised

    def is_modified(self, path_list):
        return condition.Modified(self, path_list)

    def non_existent(self, path_to_check):
        return condition.NonExistent(path_to_check)

    def rebuilt(self, build):
        return condition.Rebuilt(self, build)

    def not_run_since(self, seconds=0, minutes=0, hours=0, days=0, timedelta=None):
        """
        Returns condition that triggers when a period of time has elapsed since
        last run.

        All parameters are added together.
        - seconds, minutes, hours days are integers
        - timedelta should be a python timedelta object
        """
        td = python_timedelta()
        if timedelta is not None:
            assert isinstance(timedelta, python_timedelta), "must be python timedelta"
            td = td + timedelta
        td = td + python_timedelta(seconds=seconds)
        td = td + python_timedelta(minutes=minutes)
        td = td + python_timedelta(hours=hours)
        td = td + python_timedelta(days=days)
        return condition.NotRunSince(self, td)

    def context_manager(self):
        return BuildContextManager(self)



# -*- coding: utf-8 -*
from setuptools.command.install import install
from setuptools import find_packages
from setuptools import setup
import subprocess
import codecs
import sys
import os

def read(*parts):
    # intentionally *not* adding an encoding option to open
    # see here: https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(os.path.join(os.path.abspath(os.path.dirname(__file__)), *parts), 'r').read()

setup(name="hitchbuild",
    version=read('VERSION').replace('\n', ''),
    description="Building blocks for a build system.",
    long_description=read('README.rst'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Libraries',
        'Operating System :: Unix',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='hitchdev framework build',
    author='Colm O\'Connor',
    author_email='colm.oconnor.github@gmail.com',
    url='https://hitchbuild.readthedocs.org/',
    license='AGPL',
    install_requires=[
        'path.py', 'peewee', 'python-slugify',
    ],
    packages=find_packages(exclude=["docs", ]),
    package_data={},
    zip_safe=False,
    include_package_data=True,
)



from os import path as ospath
from path import Path
import datetime


class Change(object):
    def __nonzero__(self):
        return self.__bool__()


class YesChange(Change):
    def __init__(self, why):
        self.why = why

    def __bool__(self):
        return True


class NoChange(Change):
    def __bool__(self):
        return False


class FileChange(Change):
    def __init__(self, new, modified):
        self._new = new
        self._modified = modified

    @property
    def why(self):
        contents = ""
        if len(self._new) > 0:
            contents += "New file(s) / director(ies) detected:\n"
            for new in self._new:
                contents += "  - {0}\n".format(new)
        if len(self._modified) > 0:
            contents += "File(s) / director(ies) changed:\n"
            for modified in self._modified:
                contents += "  - {0}\n".format(modified)
        return contents

    def __bool__(self):
        return len(self._new) > 0 or len(self._modified) > 0


class Changes(object):
    def __init__(self, changes):
        self._changes = changes

    @property
    def why(self):
        return '\n'.join([
            change.why for change in self._changes
        ])

    def __bool__(self):
        return any(
            bool(change) for change in self._changes
        )


class Condition(object):
    def __init__(self):
        self._other_condition = None

    def __or__(self, other):
        self._other_condition = other
        return self

    def all_conditions(self):
        conditions = [self, ]
        other_condition = self._other_condition

        while other_condition is not None:
            conditions.append(other_condition)
            other_condition = other_condition._other_condition
        return conditions

    def check_all(self):
        return Changes([
            condition.check() for condition
            in self.all_conditions()
        ])


class Always(Condition):
    def check(self):
        return YesChange("Always triggered")


class Never(Condition):
    def check(self):
        return NoChange()


class Rebuilt(Condition):
    def __init__(self, monitor, build):
        self._monitor = monitor
        self._build = build

    def check(self):
        return self._monitor.model_for(self._build).was_triggered_on_last_run


class NonExistent(Condition):
    def __init__(self, path_to_check):
        super(NonExistent, self).__init__()
        self._path_to_check = Path(path_to_check)

    def check(self):
        return not self._path_to_check.exists()


class TimeElapsedChange(Change):
    def __init__(self, last_run, duration):
        self._last_run = last_run
        self._duration = duration

    """
    @property
    def why(self):
        return "Should run every time {0} elapses and {1} elapsed.".format(
            humanize.naturaldelta(self._duration),
            humanize.naturaldelta(datetime.datetime.now() - self._last_run),
        )
    """

    def __bool__(self):
        return True


class NotRunSince(Condition):
    def __init__(self, monitor, timedelta):
        self._monitor = monitor
        self._timedelta = timedelta
        super(NotRunSince, self).__init__()

    def check(self):
        model = self._monitor.build_model
        if model.last_run is None:
            return YesChange("Never run.")

        if model.last_run + self._timedelta < datetime.datetime.now():
            return TimeElapsedChange(model.last_run, self._timedelta)
        else:
            return NoChange()


class Modified(Condition):
    def __init__(self, monitor, paths):
        self._monitor = monitor
        self._paths = paths
        super(Modified, self).__init__()

    def check(self):
        new_files = list(self._paths)
        modified_files = []

        for monitored_file in self._monitor.File.filter(build=self._monitor.build_model):
            filename = monitored_file.filename
            if filename in self._paths:
                new_files.remove(filename)

                if ospath.getmtime(filename) > monitored_file.last_modified:
                    modified_files.append(filename)
                    monitored_file.last_modified = ospath.getmtime(filename)
                    monitored_file.save()

        for filename in new_files:
            file_model = self._monitor.File(
                build=self._monitor.build_model,
                filename=filename,
                last_modified=ospath.getmtime(filename),
            )
            file_model.save()

        return FileChange(new_files, modified_files)
